/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "CAENQdcController.h"
#include "drivers/global/DriversCommands.h"
#include "controllers/common/family/Families.h"
#include <boost/thread/thread.hpp>
#include <lstdpp128/lstdpp.h>

using namespace std;
using namespace boost;

namespace npp {

const string CAENQdcController::TYPE = "qdc_controller";

/*
 * Constructor
 */
CAENQdcController::CAENQdcController(const string& name) : CAENAdcController(name) {

	setFamily(family::ACQUISITION, family::SETTING);

	// Register properties
	triggerHoldOff.init(this, NOSAVE, "trigger_holdoff");
	chargeSensitivity.init(this, NOSAVE, "charge_sensitivity");
	gateWidth.init(this, NOSAVE, "gate_width");
	preGate.init(this, NOSAVE, "pre_gate");
	baselineMean.init(this, NOSAVE, "base_mean");
	displayGroup.init(this, SAVE, "display_group");
	channelActive_0.init(this, NOSAVE, "channel_active_0");
	channelActive_1.init(this, NOSAVE, "channel_active_1");
	channelActive_2.init(this, NOSAVE, "channel_active_2");
	channelActive_3.init(this, NOSAVE, "channel_active_3");
	channelActive_4.init(this, NOSAVE, "channel_active_4");
	channelActive_5.init(this, NOSAVE, "channel_active_5");
	channelActive_6.init(this, NOSAVE, "channel_active_6");
	channelActive_7.init(this, NOSAVE, "channel_active_7");
	triggerThreshold_0.init(this, NOSAVE, "trigger_threshold_0");
	triggerThreshold_1.init(this, NOSAVE, "trigger_threshold_1");
	triggerThreshold_2.init(this, NOSAVE, "trigger_threshold_2");
	triggerThreshold_3.init(this, NOSAVE, "trigger_threshold_3");
	triggerThreshold_4.init(this, NOSAVE, "trigger_threshold_4");
	triggerThreshold_5.init(this, NOSAVE, "trigger_threshold_5");
	triggerThreshold_6.init(this, NOSAVE, "trigger_threshold_6");
	triggerThreshold_7.init(this, NOSAVE, "trigger_threshold_7");

	m_dgtzDriver.init(this, "qdc_driver");

	registerFunction(TYPE);
}

/*
 * Destructor
 */
CAENQdcController::~CAENQdcController() {

}

/*
 * postConfiguration
 */
void CAENQdcController::postConfiguration() {

	registerRefresher(recordLength, &CAENQdcController::refreshRecordLengthProperty, this);
	registerRefresher(nbChannels, &CAENQdcController::refreshNbChannelsProperty, this);
	registerRefresher(dppAcqMode, &CAENQdcController::refreshDppAcqModeProperty, this);
	registerRefresher(displayGroup, &CAENQdcController::refreshDisplayGroupProperty, this);
	registerRefresher(triggerHoldOff, &CAENQdcController::refreshTriggerHoldOffProperty, this);
	registerRefresher(chargeSensitivity, &CAENQdcController::refreshChargeSensitivityProperty, this);
	registerRefresher(gateWidth, &CAENQdcController::refreshGateWidthProperty, this);
	registerRefresher(preGate, &CAENQdcController::refreshPreGateProperty, this);
	registerRefresher(baselineMean, &CAENQdcController::refreshBaselineMean, this);
	registerRefresher(channelActive, &CAENQdcController::refreshChannelActiveProperty, this);
	registerRefresher(dcOffset, &CAENQdcController::refreshDcOffsetProperty, this);
	registerRefresher(preTrigger, &CAENQdcController::refreshPreTriggerProperty, this);
	registerRefresher(pulsePolarity, &CAENQdcController::refreshPulsePolarityProperty, this);
	registerRefresher(channelActive_0, &CAENQdcController::refreshActivateChannelProperty, this, 0);
	registerRefresher(channelActive_1, &CAENQdcController::refreshActivateChannelProperty, this, 1);
	registerRefresher(channelActive_2, &CAENQdcController::refreshActivateChannelProperty, this, 2);
	registerRefresher(channelActive_3, &CAENQdcController::refreshActivateChannelProperty, this, 3);
	registerRefresher(channelActive_4, &CAENQdcController::refreshActivateChannelProperty, this, 4);
	registerRefresher(channelActive_5, &CAENQdcController::refreshActivateChannelProperty, this, 5);
	registerRefresher(channelActive_6, &CAENQdcController::refreshActivateChannelProperty, this, 6);
	registerRefresher(channelActive_7, &CAENQdcController::refreshActivateChannelProperty, this, 7);
	registerRefresher(triggerThreshold_0, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_1, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_2, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_3, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_4, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_5, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_6, &CAENQdcController::refreshTriggerThresholdProperty, this);
	registerRefresher(triggerThreshold_7, &CAENQdcController::refreshTriggerThresholdProperty, this);

	updateBoardProperties();
}

void CAENQdcController::refreshRecordLengthProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->recordLength = aValue;
}

void CAENQdcController::refreshDppAcqModeProperty(int32 aValue)  throw (CannotSetValue) {

	cout << "refreshDppAcqModeProperty" << endl;
	CAENAdcController::refreshDppAcqModeProperty(aValue);
	m_dgtzDriver->dppAcqMode = aValue;
	boost::thread td(boost::bind(&CAENQdcController::initBoard, this));
}

void CAENQdcController::refreshDisplayGroupProperty(int32 aGroup) throw (CannotSetValue) {

	cout << "refreshDisplayGroupProperty: " << aGroup << endl;
	float64 holdOff = m_dgtzDriver->triggerHoldOff.get(aGroup) * 16 / 1000.0;
	if (holdOff == 0)
		holdOff = 0.2;
	triggerHoldOff.update(holdOff);

	int32 trigTrsd = m_dgtzDriver->triggerThreshold.get(aGroup);
	if (trigTrsd == 0)
		trigTrsd = 20;
	triggerThreshold_0.update(trigTrsd);

	chargeSensitivity.update(m_dgtzDriver->chargeSensitivity.get(aGroup));

	float64 gWidth = m_dgtzDriver->gateWidth.get(aGroup) * 16 / 1000.0;
	if (gWidth == 0.0)
		gWidth = 1.0;
	gateWidth.update(gWidth);

	float64 pGate = m_dgtzDriver->preGate.get(aGroup) * 16 / 1000.0;
	if (pGate == 0)
		pGate = 0.1;
	preGate.update(pGate);

	int32 baseAvrg = m_dgtzDriver->baselineMean.get(aGroup);
	if (baseAvrg == 0)
		baseAvrg = 4;
	baselineMean.update(baseAvrg);

	//channelActive.update(m_dgtzDriver->channelActive.get(aGroup));

	int32 offSet = m_dgtzDriver->dcOffset.get(aGroup);
	if (offSet == 0)
		offSet = 10;
	dcOffset.update(offSet);

	int32 preTrig = m_dgtzDriver->preTrigger.get(aGroup);
	if (preTrig == 0)
		preTrig = 500;
	preTrigger.update(preTrig);

	pulsePolarity.update(m_dgtzDriver->pulsePolarity.get(aGroup));
}

void CAENQdcController::refreshTriggerHoldOffProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerHoldOff.set(displayGroup(), (int32)(aValue * 1000 / 16));
	cout << "refreshTriggerHoldOffProperty: " << (int32)(aValue * 1000 / 16) << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshTriggerThresholdProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->triggerThreshold.set(displayGroup(), aValue);
	cout << "refreshTriggerThresholdProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshChargeSensitivityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->chargeSensitivity.set(displayGroup(), aValue);
	cout << "refreshChargeSensitivityProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshGateWidthProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->gateWidth.set(displayGroup(), (int32)(aValue * 1000 / 16));
	cout << "refreshGateWidthProperty: " << (int32)(aValue * 1000 / 16) << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPreGateProperty(float64 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preGate.set(displayGroup(), (int32)(aValue * 1000 / 16));
	cout << "refreshPreGateProperty: " << (int32)(aValue * 1000 / 16) << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshBaselineMean(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->baselineMean.set(displayGroup(), aValue);
	cout << "refreshBaselineMean: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshChannelActiveProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->channelActive.set(displayGroup(), aValue);
	cout << "refreshChannelActiveProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshDcOffsetProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->dcOffset.set(displayGroup(), aValue);
	cout << "refreshDcOffsetProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPreTriggerProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->preTrigger.set(displayGroup(), aValue);
	cout << "refreshPreTriggerProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshPulsePolarityProperty(int32 aValue) throw (CannotSetValue) {

	m_dgtzDriver->pulsePolarity.set(displayGroup(), aValue);
	cout << "refreshPulsePolarityProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::refreshActivateChannelProperty(int32 index, int32 aValue) throw (CannotSetValue) {

	int32 aChannel = displayGroup() * 8 + index + 1;
	m_dgtzDriver->channelActive.set(aChannel, aValue);
	cout << "refreshActivateChannelProperty: " << aValue << endl;
	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
}

void CAENQdcController::updateBoardProperties() {

	modelName.update(m_dgtzDriver->modelName());
	channels.update(m_dgtzDriver->channels());
	nbChannels.update(m_dgtzDriver->nbChannels());
	minRange = 0;
	maxRange = m_dgtzDriver->nbChannels() - 1;
	dppAcqMode.update(m_dgtzDriver->dppAcqMode());
	cardType = lstdpp128::C1740_QDC;
}

void CAENQdcController::writeParam() {

	m_dgtzDriver.execute(caen_qdc::CAENQdcDriver::WRITE_PARAMETERS_COMMAND, false);
}

void CAENQdcController::initBoard() {

	m_dgtzDriver.execute(driver::INIT_COMMAND, false);
}

void CAENQdcController::refreshNbChannelsProperty(int32 aGroup) throw (CannotSetValue) {

}

void CAENQdcController::refreshDisplayChannelProperty(int32 aGroup) throw (CannotSetValue)  {

}

void CAENQdcController::initChannel(int32 aGroup) {
	/*

	m_dgtzDriver->triggerHoldOff.update((int32)(triggerHoldOff() * 1000));
	m_dgtzDriver->triggerThreshold.set(aChannel, triggerThreshold());
	m_dgtzDriver->selfTrigger.set(aChannel,selfTrigger());
	m_dgtzDriver->chargeSensitivity.set(aChannel, chargeSensitivity());
	m_dgtzDriver->preGate.set(aChannel, (int32)(preGate() * 1000));
	m_dgtzDriver->gateWidth.set(aChannel, (int32)(gateWidth() * 1000));
	m_dgtzDriver->preGate.set(aChannel, (int32)(preGate() * 1000));
	m_dgtzDriver->baselineMean.set(aChannel, baselineMean());
	m_dgtzDriver->channelActive.set(aChannel, channelActive());
	m_dgtzDriver->dcOffset.set(aChannel, dcOffset());
	m_dgtzDriver->preTrigger.set(aChannel, preTrigger());
	m_dgtzDriver->pulsePolarity.set(aChannel, pulsePolarity());

	// Update channel mask
	int32 mask = 0;
	for (int32 i=0; i<m_dgtzDriver->nbChannels(); ++i) {
		if (m_dgtzDriver->channelActive.get(i))
				mask += (m_dgtzDriver->channelActive.get(i) << i);
	}
	m_dgtzDriver->channelMask = mask;

	m_dgtzDriver->dcOffset.set(aChannel, dcOffset());
	m_dgtzDriver->preTrigger.set(aChannel, preTrigger());
	m_dgtzDriver->pulsePolarity.set(aChannel, pulsePolarity());

	boost::thread td(boost::bind(&CAENQdcController::writeParam, this));
*/
}

void CAENQdcController::clearData() {
	commandProgression = 0;
	m_dgtzDriver.execute(caen_qdc::CAENQdcDriver::CLEAR_COMMAND, true);
	commandProgression = 100;
}

}
