/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealCAENPhaDriver.h"

#include <controllers/common/acquisition/mode/DPPAcquisitionCommon.h>
#include "CAENPhaDriver.h"

using namespace std;

namespace caen_pha {

/*
 * Constructor
 */
RealCAENPhaDriver::RealCAENPhaDriver(CAENPhaDriver* owner) :
	CAENPhaState(owner) {

	m_boardHandle = 0;
	ret = CAEN_DGTZ_Success;
}

/*
 * Destructor
 */
RealCAENPhaDriver::~RealCAENPhaDriver() {

}

/*!
 * \brief open command implementation
 */
void RealCAENPhaDriver::open() {

	boardParams.LinkType = (CAEN_DGTZ_ConnectionType)owner()->linkType();
	boardParams.VMEBaseAddress = owner()->address();
	ret = CAEN_DGTZ_OpenDigitizer(boardParams.LinkType, 0, 0, boardParams.VMEBaseAddress, &m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_OpenDigitizer: " << ret << endl;
		return;
	}

}

/*!
 * \brief close command implementation
 */
void RealCAENPhaDriver::close() {

	CAEN_DGTZ_CloseDigitizer(m_boardHandle);
}

/*!
 * \brief Init command implementation
 */
void RealCAENPhaDriver::init() {

	reset();

	readInfos();

	open();

    // Channel CTRL register (individual trigger, sequencial readout)
	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x01000114);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}

	// Set acquisition mode
	if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_List;
	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Oscilloscope;
	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::WAVEFORM_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Oscilloscope;
	}
	boardParams.SaveMode = (CAEN_DGTZ_DPP_SaveParam_t)owner()->saveMode();
	if (owner()->modelName() == "Pha") {
		ret = CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, boardParams.SaveMode);
	} else {
		if (boardParams.DPPAcqMode == 0)
			boardParams.DPPAcqMode = (CAEN_DGTZ_DPP_AcqMode_t)2;
		ret = CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, boardParams.SaveMode);
	}
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPPAcquisitionMode: " << ret << endl;
		close();
		return;
	}

	// Set start acquisition mode
	boardParams.AcqMode = (CAEN_DGTZ_AcqMode_t)owner()->acqMode();
	ret = CAEN_DGTZ_SetAcquisitionMode(m_boardHandle, CAEN_DGTZ_S_IN_CONTROLLED); //boardParams.AcqMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetAcquisitionMode: " << ret << endl;
		close();
		return;
	}

	// Set nb of samples for oscilloscope mode
	boardParams.RecordLength = owner()->recordLength();
	ret = CAEN_DGTZ_SetRecordLength(m_boardHandle, boardParams.RecordLength);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetRecordLength: " << ret << endl;
		close();
		return;
	}

	// Set I/O level
	boardParams.IOLevel = (CAEN_DGTZ_IOLevel_t)owner()->ioLevel();
	ret = CAEN_DGTZ_SetIOLevel(m_boardHandle, boardParams.IOLevel);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetIOLevel: " << ret << endl;
		close();
		return;
	}

	// Set external trigger
	boardParams.TriggerMode = (CAEN_DGTZ_TriggerMode_t)owner()->triggerMode();
	ret = CAEN_DGTZ_SetExtTriggerInputMode(m_boardHandle, boardParams.TriggerMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetExtTriggerInputMode: " << ret << endl;
		close();
		return;
	}

	// Enable channels
	boardParams.ChannelMask = owner()->channelMask();
	ret = CAEN_DGTZ_SetChannelEnableMask(m_boardHandle, boardParams.ChannelMask);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetChannelEnableMask: " << ret << endl;
		close();
		return;
	}

	// Set aggregate size - how many events per output buffer
	// 0 = automatic
	boardParams.EventAggr = 0;
	ret = CAEN_DGTZ_SetDPPEventAggregation(m_boardHandle, boardParams.EventAggr, 0);
	if (ret) {
		cout << "1724 error" << endl;
		cerr << "Error CAEN_DGTZ_SetDPPEventAggregation: " << ret << endl;
		close();
		return;
	}

	// Set multi-board synchronization mode
	ret = CAEN_DGTZ_SetRunSynchronizationMode(m_boardHandle, CAEN_DGTZ_RUN_SYNC_Disabled);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetRunSynchronizationMode: " << ret << endl;
		close();
		return;
	}

	// Set virtual probe settings
	if (owner()->modelName() == "V1724") {
		ret = CAEN_DGTZ_SetDPP_PHA_VirtualProbe(m_boardHandle, CAEN_DGTZ_DPP_VIRTUALPROBE_DUAL,
				CAEN_DGTZ_DPP_PHA_VIRTUALPROBE1_trapezoid, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE2_Input,
				CAEN_DGTZ_DPP_PHA_DIGITAL_PROBE_Peaking);
	} else {
	    ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_1, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE1_trapezoid);
	    ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_2, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE2_Input);
	    ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, DIGITAL_TRACE_1, CAEN_DGTZ_DPP_DIGITALPROBE_Peaking);
//		ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_1, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE1_trapezoid);
//		ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_2, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE2_Input);
//		ret = CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, DIGITAL_TRACE_3, CAEN_DGTZ_DPP_PHA_DIGITAL_PROBE_Peaking);
	}
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPP_VirtualProbe: " << ret << endl;
		close();
		return;
	}

	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8100, 0x5);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}

	// PN1 COINCIDENCE MODE =====> HACK EMILIO
	if (owner()->usePn1Coincidence() == 1) {
		// Channel CTRL register (individual trigger, sequencial readout)
		ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x01000114);
		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
			close();
			return;
		}
	}
//	// HACK EMILIO VIVALDI 20/10/2017
//	// waveform mode -- trig-IN
//
//	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x01070914);// VIVALDI disable dual trace
//			if (ret) {
//				cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
//				close();
//				return;
//			}
//	//-- DISABLE EXTERNAL TRIGGER 0x817C
//	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x817C, 0x0);
//		if (ret) {
//			cerr << "Error CAEN_DGTZ_WriteRegister Disable External Trigger: " << ret << endl;
//			close();
//			return;
//		}
//
//	//-- FRONT PANEL I/O CONTROL 0x811C
//	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x811C, 0x801);
//		if (ret) {
//			cerr << "Error CAEN_DGTZ_WriteRegister FRONT PANEL I/O CONTROL: " << ret << endl;
//			close();
//			return;
//		}
//
//	//-- GLOBAL TRIGGER MASK 0x810C
//	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x810C, 0x40000000);
//		if (ret) {
//			cerr << "Error CAEN_DGTZ_WriteRegisterGLOBAL TRIGGER MASK: " << ret << endl;
//			close();
//			return;
//		}
//
//	//--  TRIGGER VALIDATION MASK 0x8180
//	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8180, 0x40000000);
//	if (ret) {
//		cerr << "Error CAEN_DGTZ_WriteRegisterGLOBAL TRIGGER MASK: " << ret << endl;
//		close();
//		return;
//	}
	//
	close();

	writeParam();
}

/*!
 * \brief Clear command implementation
 */
void RealCAENPhaDriver::clear() {

	open();
	boardParams.IOLevel = (CAEN_DGTZ_IOLevel_t)owner()->ioLevel();
		ret = CAEN_DGTZ_SetIOLevel(m_boardHandle, boardParams.IOLevel);
		if (ret) {
			cerr << "Error CAEN_DGTZ_SetIOLevel: " << ret << endl;
			close();
			return;
		}
	/*ret = CAEN_DGTZ_ClearData(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_ClearData: " << ret << endl;
	}*/
	close();
}

/*!
 * \brief Reset command implementation
 */
void RealCAENPhaDriver::reset() {

	open();
	ret = CAEN_DGTZ_Reset(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_Reset: " << ret << endl;
	}
	close();
}

/*!
 * \brief Write Param command implementation
 */
void RealCAENPhaDriver::writeParam() {

	// Load DPP parameters into structure
	for (int32 i=0; i<owner()->nbChannels(); ++i) {
		dppParams.M[i] = owner()->signalDecayTime.get(i);
		dppParams.m[i] = owner()->trapezoidFlatTop.get(i);
		dppParams.k[i] = owner()->trapezoidRiseTime.get(i);
		dppParams.ftd[i] = owner()->flatTopDelay.get(i);
		dppParams.a[i] = owner()->triggerSmoothing.get(i);
		dppParams.b[i] = owner()->signalRiseTime.get(i);
		dppParams.thr[i] = owner()->triggerThreshold.get(i);
		dppParams.nsbl[i] = owner()->baselineMean.get(i);
		dppParams.nspk[i] = owner()->peakMean.get(i);
		dppParams.pkho[i] = owner()->peakHoldOff.get(i);
		dppParams.blho[i] = owner()->baselineHoldOff.get(i);
		dppParams.otrej[i] = 0;
		dppParams.trgho[i] = owner()->triggerHoldOff.get(i);
		dppParams.twwdt[i] = owner()->triggerWindowSize.get(i);
		dppParams.trgwin[i] = owner()->triggerWindowEnable.get(i);
		dppParams.dgain[i] = owner()->digitalGain.get(i);
		dppParams.enf[i] = owner()->trapezoidGain.get(i);
		dppParams.decimation[i] = owner()->decimation.get(i);
	}

	open();

	// Write channels parameters
	ret = CAEN_DGTZ_SetDPPParameters(m_boardHandle, boardParams.ChannelMask, &dppParams);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPPParameters" << endl;
		close();
		return;
	}

	for (int32 i=0; i<owner()->channels(); ++i) {
		if (owner()->channelMask() & (1 << i)) {
			// Set DC offset
			int32 offset = owner()->dcOffset.get(i);
			offset +=50;
			offset = offset * 65535 / 100;
			ret = CAEN_DGTZ_SetChannelDCOffset(m_boardHandle, i, offset);
			if (ret) {
				cerr << "Error CAEN_DGTZ_SetChannelDCOffset" << endl;
				close();
				return;
			}

			// Set Pre-trigger size
			int32 ptgrSize = owner()->preTrigger.get(i);
			ret = CAEN_DGTZ_SetDPPPreTriggerSize(m_boardHandle, i, ptgrSize);
			if (ret) {
				cerr << "Error CAEN_DGTZ_SetDPPPreTriggerSize" << endl;
				close();
				return;
			}

			// Set polarity
			CAEN_DGTZ_PulsePolarity_t polarity = (CAEN_DGTZ_PulsePolarity_t)owner()->pulsePolarity.get(i);
			ret = CAEN_DGTZ_SetChannelPulsePolarity(m_boardHandle, i, polarity);
			if (ret) {
				cerr << "Error CAEN_DGTZ_SetChannelPulsePolarity" << endl;
				close();
				return;
			}

		    // Enable Fake Events for TimeTag Rollover (used for low rates applications)
			uint32_t val;
            uint32_t addr = PHA_FAKEEVTS_ROLLOVER_REG | (i<<8);
            //uint32_t bitMask = 0x04000000; // bit 26
            uint32_t bitMask = 0xFBFFFFFF; // bit 26 Not used in PF1 -- HACk Emilio IFC1210 does the roll over
            ret = CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
			if (ret) {
				cerr << "Error CAEN_DGTZ_ReadRegister" << endl;
				close();
				return;
			}
            //val |= bitMask;
            val &= bitMask;
            ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
			if (ret) {
				cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
				close();
				return;
			}


			// Disable SElf TRigger HACK VIVALDI
//						//uint32_t val;
//			            //uint32_t addr = PHA_FAKEEVTS_ROLLOVER_REG | (i<<8);
//			            bitMask = 0x01000000; // bit 24
//
//			            ret = CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
//						if (ret) {
//							cerr << "Error CAEN_DGTZ_ReadRegister" << endl;
//							close();
//							return;
//						}
//			            val |= bitMask;
//			            //val &= bitMask;
//			            ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
//						if (ret) {
//							cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
//							close();
//							return;
//						}
		}
	}

	close();
}

/*!
 * \brief Set acquisition mode command implementation
 */
void RealCAENPhaDriver::setAcqMode() {

	open();
	boardParams.DPPAcqMode = (CAEN_DGTZ_DPP_AcqMode_t)owner()->dppAcqMode();
	boardParams.SaveMode = (CAEN_DGTZ_DPP_SaveParam_t)owner()->saveMode();
	ret = CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, boardParams.SaveMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPPAcquisitionMode: " << ret << endl;
	}
	close();
}

/*!
 * \brief Read Infos command implementation
 */
void RealCAENPhaDriver::readInfos() {

	open();
	ret = CAEN_DGTZ_GetInfo(m_boardHandle, &boardInfo);
	close();
	owner()->nbBits = boardInfo.ADC_NBits;
	owner()->model = boardInfo.Model;
	owner()->channels = boardInfo.Channels;
	owner()->serialNumber = boardInfo.SerialNumber;
	owner()->modelName = boardInfo.ModelName;
	owner()->rocFirmware = boardInfo.ROC_FirmwareRel;
	owner()->amcFirmware = boardInfo.AMC_FirmwareRel;
}

/*!
 * \brief Enable Coincidence command implementation
 */
void RealCAENPhaDriver::useCoincidence() {

	open();
	uint32_t val;
    uint32_t addr = 0x1180;
    ret = CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
	if (ret) {
		cerr << "Error CAEN_DGTZ_ReadRegister" << endl;
		close();
		return;
	}
	if (owner()->enableCoincidence() == 0) {
		val = val & 0x80000;
	} else {
		val = val | (1 << 19);
	}
    ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
		close();
		return;
	}

	if (owner()->enableCoincidence() == 1) {
		val = owner()->coincidenceWindow();
		addr = 0x1184;
		ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
			close();
			return;
		}
	    uint32_t addr = 0x8184;
	    ret = CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
		if (ret) {
			cerr << "Error CAEN_DGTZ_ReadRegister" << endl;
			close();
			return;
		}
		val = (val & 0x3FF) + owner()->coincidenceMask();
	    ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
			close();
			return;
		}
	}

	close();
}

}
