/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef CAENQDCDRIVER_H
#define CAENQDCDRIVER_H

#include <Driver.h>
#include "drivers/global/Vme.h"

namespace caen_qdc {

/*!
 * \class CAENQdcDriver
 * \brief Main class for the CAENQdcDriver device driver
 *
 * \par
 * Class for CAENQdcDriver driver. Implements DeviceDriver methods.
 * \par
 * The CAENQdcDriver device manages the global functionality of the CAEN V1724 VME Digitiser.
 */

class CAENQdcDriver : public driver::Vme {

public:

	//! Driver type value
	static const std::string TYPE;

	/*!
	 * \brief Constructor
	 * \param[in] name the name of the device driver
	 */
	CAENQdcDriver(const std::string& name);

	/*!
	 * \brief Destructor
	 */
	virtual ~CAENQdcDriver();

	/*!
	 * \brief Method called for executing a command
	 *
	 * \param[in] command the command to apply on the controller
	 */
	virtual void execute(const std::string& aCommand);

	// Global properties
	Property<int32> nbChannels;
	Property<int32> nbBits;
    Property<int32> model;
    Property<int32>	channels;
    Property<int32> serialNumber;
    Property<std::string> modelName;
    Property<std::string> rocFirmware;
    Property<std::string> amcFirmware;

    // Board properties
    Property<int32> linkType;
    Property<int32> acqMode;
    Property<int32> dppAcqMode;
    Property<std::string> dppAcqModeStr;
    Property<int32> saveMode;
    Property<std::string> saveModeStr;
    Property<int32> recordLength;					// Nb samples for oscilloscope mode
    Property<int32> ioLevel;
    Property<int32> triggerMode;

	// Group properties in dppParams - QDC
    DynamicProperty<int32> gateWidth;				// sgate - samples (16 ns)
    DynamicProperty<int32> preGate;					// pgate - samples (16 ns)
    DynamicProperty<int32> baselineMean;			// nsbl - range 0:6 (no baseline, 4, 8, 16, 32, 64, 128)
	DynamicProperty<int32> preTrigger;				// samples - must be > preGate + 112 ns
    DynamicProperty<int32> triggerHoldOff;			// trgho - samples
    DynamicProperty<int32> triggerHysteresis;		// range 0:1 enable, disable
	DynamicProperty<int32> dcOffset;				// LSB - value x 16
	DynamicProperty<int32> channelEnableMask;		// max value 0xFF
    DynamicProperty<int32> triggerThreshold;		// thr (N x 4 ns) ---> NB per channel
    DynamicProperty<int32> chargeSensitivity;		// values 0 to 7
    DynamicProperty<int32> inputSmoothing;			// values 0 to 7
    DynamicProperty<int32> channelActive;
	DynamicProperty<int32> pulsePolarity;			// 0 = positive - 1 = negative
    DynamicProperty<std::string> channelActiveStr;
	DynamicProperty<std::string> pulsePolarityStr;

	// Commands
	static const std::string WRITE_PARAMETERS_COMMAND;
	static const std::string SET_MODE_COMMAND;
	static const std::string RESET_COMMAND;
	static const std::string CLEAR_COMMAND;

private:

	void refreshNbChannelsProperty(int32 aValue) throw (CannotSetValue);
	void refreshDppAcqModeProperty(int32 aValue) throw (CannotSetValue);
	void refreshSaveModeProperty(int32 aValue) throw (CannotSetValue);
	void refreshChannelActiveProperty(int32 index, int32 aValue) throw (CannotSetValue);
	void refreshPulsePolarityProperty(int32 index, int32 aValue) throw (CannotSetValue);
};

}
#endif //CAENQDCDRIVER_H
