/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealCAENQdcDriver.h"

#include <controllers/common/acquisition/mode/DPPAcquisitionCommon.h>
#include "CAENQdcDriver.h"

using namespace std;

namespace caen_qdc {

/*
 * Constructor
 */
RealCAENQdcDriver::RealCAENQdcDriver(CAENQdcDriver* owner) : CAENQdcState(owner) {

	m_boardHandle = 0;
	ret = CAEN_DGTZ_Success;
}

/*
 * Destructor
 */
RealCAENQdcDriver::~RealCAENQdcDriver() {

}

/*!
 * \brief open command implementation
 */
void RealCAENQdcDriver::open() {

	boardParams.LinkType = (CAEN_DGTZ_ConnectionType)owner()->linkType();
	boardParams.VMEBaseAddress = owner()->address();
	ret = CAEN_DGTZ_OpenDigitizer(boardParams.LinkType, 0, 0, boardParams.VMEBaseAddress, &m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_OpenDigitizer: " << ret << endl;
		return;
	}
}

/*!
 * \brief close command implementation
 */
void RealCAENQdcDriver::close() {

	CAEN_DGTZ_CloseDigitizer(m_boardHandle);
}

/*!
 * \brief Init command implementation
 */
void RealCAENQdcDriver::init() {

	reset();

	readInfos();

	open();

	// Channel CTRL register (individual trigger, sequencial readout)
	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x003E1110);
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 0x00300000);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}

	// Set acquisition mode
	if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_List;
	}
	else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Oscilloscope;
	}
	else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::WAVEFORM_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Oscilloscope;
	}

/*	boardParams.SaveMode = (CAEN_DGTZ_DPP_SaveParam_t)owner()->saveMode();
	ret = CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, boardParams.SaveMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPPAcquisitionMode: " << ret << endl;
		close();
		return;
	}
*/
	// Set start acquisition mode
	boardParams.AcqMode = (CAEN_DGTZ_AcqMode_t)owner()->acqMode();
	ret = CAEN_DGTZ_SetAcquisitionMode(m_boardHandle, CAEN_DGTZ_S_IN_CONTROLLED); //boardParams.AcqMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetAcquisitionMode: " << ret << endl;
		close();
		return;
	}

	// Set I/O level
	boardParams.IOLevel = (CAEN_DGTZ_IOLevel_t)owner()->ioLevel();
	ret = CAEN_DGTZ_SetIOLevel(m_boardHandle, boardParams.IOLevel);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetIOLevel: " << ret << endl;
		close();
		return;
	}

	// Set external trigger
	boardParams.TriggerMode = (CAEN_DGTZ_TriggerMode_t)owner()->triggerMode();
	ret = CAEN_DGTZ_SetExtTriggerInputMode(m_boardHandle, boardParams.TriggerMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetExtTriggerInputMode: " << ret << endl;
		close();
		return;
	}

    /* Trigger Hold Off */
    //ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8074, 10);
	//if (ret) {
	//	cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
	//	close();
	//	return;
	//}

	// Enable channels
	//boardParams.ChannelMask = owner()->channelMask();
	//ret = CAEN_DGTZ_SetChannelEnableMask(m_boardHandle, boardParams.ChannelMask);
	//ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x80A8, 0XFF);
	//if (ret) {
	//	cerr << "Error CAEN_DGTZ_SetChannelEnableMask: " << ret << endl;
	//	close();
	//	return;
	//}

	//uint32_t DppCtrl1 = 0;
	//DppCtrl1 = (((owner()->triggerHysteresis.get(0) & 0x1)  << 30)   |
	//		((owner()->selfTrigger.get(0) & 1)     			<< 24)   |
	//		((owner()->baselineMean.get(0) & 0x7)     		<< 20)   |
	//		((owner()->triggerMode() & 3)            		<< 18)   |
	//		((owner()->chargeSensitivity.get(0) & 0x7)		<<  0)   |
	//		((owner()->pulsePolarity.get(0) & 1)           	<< 16)   |
	//		((owner()->inputSmoothing.get(0) & 7)       	<< 12));
	//ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8040, DppCtrl1);
	//if (ret) {
	//	cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
	//	close();
	//	return;
	//}

    /* Set Pre Trigger (in samples) */
	//ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x803C, 20);

    /* Set Gate Offset (in samples) */
	//ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8034, 0);

    /* Set Baseline (used in fixed baseline mode only) */
	//ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8038, 2100);

    /* Set Gate Width (in samples) */
    //for(int32_t i=0; i<8; i++) {
    //	 ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1030 + 0x100*i, 100);
    //}

    // Set aggregate size - how many events per output buffer
	// 0 = automatic
//	if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
//		boardParams.EventAggr = 0;
//        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);	// Buffer organization
//	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
//		boardParams.EventAggr = 16;
//        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0xA);
//	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::WAVEFORM_MODE) {
//		boardParams.EventAggr = 1;
//        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);	// Buffer organization
//	}

    // Set aggregate size - how many events per output buffer
	ret = 0;
    //ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 0x20);
	//ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0xA);
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0xEF1C, 0x1);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}

	// Set multi-board synchronization mode
	ret = CAEN_DGTZ_SetRunSynchronizationMode(m_boardHandle, CAEN_DGTZ_RUN_SYNC_Disabled);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetRunSynchronizationMode: " << ret << endl;
		close();
		return;
	}
/*
	// Set virtual probe settings
	ret = CAEN_DGTZ_SetDPP_PSD_VirtualProbe(m_boardHandle, CAEN_DGTZ_DPP_VIRTUALPROBE_SINGLE,
			CAEN_DGTZ_DPP_PSD_VIRTUALPROBE_Baseline, CAEN_DGTZ_DPP_PSD_DIGITALPROBE1_R6_GateLong,
			CAEN_DGTZ_DPP_PSD_DIGITALPROBE2_R6_GateShort);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPP_VirtualProbe: " << ret << endl;
		close();
		return;
	}
*/
	uint32 data;
	ret = CAEN_DGTZ_ReadRegister(m_boardHandle, 0x8000, &data);

	cout << "Acq mode: " << owner()->dppAcqMode() << endl;

	if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
		data = 0x000E0110;
		ret = 0;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0xA);
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 16);
		if (ret) {
				cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
				close();
				return;
			}
	}
	else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
		data = 0x000F0110;
		ret = 0;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 0x1);
		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
			close();
			return;
		}
	}
	else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::WAVEFORM_MODE) {
		data = 0x000F0110;
		ret = 0;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 0x1);
		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
			close();
			return;
		}
	}
	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, data);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}
	ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8100, 0x5);
	if (ret) {
		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
		close();
		return;
	}


	//for(int32_t i=0; i<1; i++) {
	//			int32_t address = 0x1024 + 0x100*i;
				//cout << "RecordLength: 0x" << hex << address << " value: 0x" << owner()->recordLength()/8 << endl;
				//ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->recordLength()/8);
	//			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, 0x50);
	//		}

	//ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8024, owner()->recordLength()/8);
	//ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8024, 0x50);
	//	if (ret) {
	//		cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
	//		close();
	//		return;
	//	}

	close();

	writeParam();
}

/*!
 * \brief Clear command implementation
 */
void RealCAENQdcDriver::clear() {

	open();
	ret = CAEN_DGTZ_ClearData(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_ClearData: " << ret << endl;
	}
	close();
}

/*!
 * \brief Reset command implementation
 */
void RealCAENQdcDriver::reset() {

	open();
	ret = CAEN_DGTZ_Reset(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_Reset: " << ret << endl;
	}
	close();
}

/*!
 * \brief Write Param command implementation
 */
void RealCAENQdcDriver::writeParam() {

	ret = 0;
	int i, j;
	int nbGroup = owner()->nbChannels() / 8;
	uint32_t DppCtrl1 = 0;

	open();

	for (i=0; i<nbGroup; i++) {
		uint32_t groupMask = 0;
		bool groupActive = false;
		for (j=0; j<8; j++) {
			int ch = i * 8 + j;
			if (owner()->channelActive.get(ch))
				groupActive = true;
		}
        if (groupActive)
			groupMask |= (1<<j);
    	owner()->channelEnableMask.set(i, groupMask);
		int32_t address = 0x10A8 + 0x100 * i;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, groupMask);
    	cout << "Group mask: 0x" << hex << groupMask << dec << endl;
	}

	/*
	 ** Set selfTrigger threshold
	 */
	for (i=0; i<64; i++) {
		ret |= setChannelTriggerThreshold(i, owner()->triggerThreshold.get(i));
	}

	/* Disable Group self trigger for the acquisition (mask = 0) */
	ret |= CAEN_DGTZ_SetGroupSelfTrigger(m_boardHandle, CAEN_DGTZ_TRGMODE_ACQ_ONLY, 0x00);

	/* Set the behaviour when a SW trigger arrives */
	ret |= CAEN_DGTZ_SetSWTriggerMode(m_boardHandle, CAEN_DGTZ_TRGMODE_ACQ_ONLY);

	/* Set the max number of events/aggregates to transfer in a single readout */
	ret |= CAEN_DGTZ_SetMaxNumAggregatesBLT(m_boardHandle, MAX_AGGR_NUM_PER_BLOCK_TRANSFER);

	/* Trigger Hold Off */
	for (i=0; i<nbGroup; ++i) {
		int32_t address = 0x1074 + 0x100 * i;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->triggerHoldOff.get(i));
	}

	/* DPP Control 1 register */
	int32_t selfTrigger = 0;
	for (i=0; i<nbGroup; ++i) {
		DppCtrl1 = (((owner()->triggerHysteresis.get(i) & 0x1)  << 30)   |
				selfTrigger						     			<< 24)   |
				((owner()->baselineMean.get(i) & 0x7)     		<< 20)   |
				((owner()->triggerMode() & 3)            		<< 18)   |
				((owner()->chargeSensitivity.get(i) & 0x7)		<<  0)   |
				((owner()->pulsePolarity.get(i) & 1)           	<< 16)   |
				((owner()->inputSmoothing.get(i) & 7)       	<< 12);

		int32_t address = 0x1040 + 0x100 * i;
		//cout << "ctrl register: 0x" << hex << address << " value: 0x" << DppCtrl1 << endl;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, DppCtrl1);
	}

	/* Set Pre Trigger (in samples) */
	for(i=0; i<nbGroup; i++) {
		int32_t address = 0x103C + 0x100 * i;
		//cout << "preTrigger: 0x" << hex << address << " value: 0x" << owner()->preTrigger.get(i) << endl;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->preTrigger.get(i));
	}

	/* Set Gate Offset (in samples) */
	for(i=0; i<nbGroup; i++) {
		int32_t address = 0x1034 + 0x100 * i;
		//cout << "preGate: 0x" << hex << address << " value: 0x" << owner()->preGate.get(i) << endl;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->preGate.get(i));
	}

	/* Set Gate Width (in samples) */
	for(i=0; i<nbGroup; i++) {
		int32_t address = 0x1030 + 0x100 * i;
		cout << "gateWidth: 0x" << hex << address << " value: 0x" << owner()->gateWidth.get(i) << endl;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->gateWidth.get(i));
	}

	/* Set the waveform lenght (in samples) */
	//ret |= setRecordLength(owner()->recordLength());
	for(i=0; i<nbGroup; i++) {
			int32_t address = 0x1024 + 0x100 * i;
			//cout << "RecordLength: 0x" << hex << address << " value: 0x" << owner()->recordLength()/8 << endl;
			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, owner()->recordLength());
		}

		/* Set DC offset */
	for (i=0; i<nbGroup; i++) {
		ret |= CAEN_DGTZ_SetGroupDCOffset(m_boardHandle, i, owner()->dcOffset.get(i));
	}

	/* Set number of events per memory buffer */
	uint32_t nEvAggr = 32;
	setNumEvAggregate(nEvAggr);

	/* Set Extended Time Stamp if enabled*/
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 1 << 17);

	/* Check errors */
	if(ret != CAEN_DGTZ_Success) {
		cout << "Errors during digitizer configuration" << endl;
	} else {
		cout << "Digitizer configuration ok" << endl;
	}

	close();
}

/*!
 * \brief Set acquisition mode command implementation
 */
void RealCAENQdcDriver::setAcqMode() {

	open();

	boardParams.DPPAcqMode = (CAEN_DGTZ_DPP_AcqMode_t)owner()->dppAcqMode();
	boardParams.SaveMode = (CAEN_DGTZ_DPP_SaveParam_t)owner()->saveMode();
	ret = CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, boardParams.SaveMode);
	if (ret) {
		cerr << "Error CAEN_DGTZ_SetDPPAcquisitionMode: " << ret << endl;
	}

	close();
}

/*!
 * \brief Read Infos command implementation
 */
void RealCAENQdcDriver::readInfos() {

	open();

	ret = CAEN_DGTZ_GetInfo(m_boardHandle, &boardInfo);

	close();

	owner()->nbBits = boardInfo.ADC_NBits;
	owner()->model = boardInfo.Model;
	owner()->channels = boardInfo.Channels * 8; // nb of group channels with 8 channels per group
	owner()->serialNumber = boardInfo.SerialNumber;
	owner()->modelName = boardInfo.ModelName;
	owner()->rocFirmware = boardInfo.ROC_FirmwareRel;
	owner()->amcFirmware = boardInfo.AMC_FirmwareRel;
}

uint32_t RealCAENQdcDriver::getChannelAddress(uint32_t base, uint16_t channel) {

	//return base + (0x100 * channel);
    int chShift = 8;
    uint32_t res = base & ~(0xF<<chShift); // set bits [11:8] to zero
    res |= channel << chShift;
    return res;
}

uint32_t RealCAENQdcDriver::setChannelTriggerThreshold(uint32_t channel, uint32_t Tvalue) {

	int gr;
    int ch;

    ret=0;
    uint32_t addr;

    if (channel > 64)
        return -1;

    gr = channel/8;
    ch = channel % 8;

    addr = 0x1000 + 0x100*gr + 0xD0 + 4*ch;
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, Tvalue);

    return ret;
}

uint32_t RealCAENQdcDriver::setRecordLength(uint32_t recordLength) {

	ret=0;

	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8024, recordLength/8);

    return ret;
}

uint32_t RealCAENQdcDriver::setNumEvAggregate(uint32_t nEvAggr) {

	ret=0;

	uint32_t d32;

    if (nEvAggr == 0) {
        CAEN_DGTZ_ReadRegister(m_boardHandle, 0x8000, &d32);
        if (d32 & (1<<16)) {                                          /* waveform enabled      */
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);      /* Buffer organization   *///0x3
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 1);        /* Events per aggregate  */
        } else {
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0xA);      /* Buffer organization   */
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 16);       /* Events per aggregate  */
        }
    } else {
        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);          /* Buffer organization   */
        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, nEvAggr);      /* Events per aggregate  */
    }

    return ret;
}

}
